@@ -245,13 +245,7 @@ enum STATFLAG {
 #define STATLEN (1 << 4)
 
 enum WB_FLAG {
-	/*
-	 * This flag is set when either one of the underlying devices returned
-	 * EIO and we must immediately block up the whole to avoid further
-	 * damage.
-	 */
-	WB_DEAD = 0,
-	WB_CREATED = 1,
+	WB_CREATED = 0,
 };
 
 #define SEGMENT_SIZE_ORDER 10
@@ -266,7 +260,7 @@ struct wb_device {
 	struct dm_dev *backing_dev; /* Slow device (HDD) */
 	struct dm_dev *cache_dev; /* Fast device (SSD) */
 
-	bool write_through_mode;
+	bool write_around_mode;
 
 	unsigned nr_ctr_args;
 	const char **ctr_args;
@@ -409,6 +403,7 @@ struct wb_device {
 
 	u32 nr_cur_batched_writeback;
 	u32 nr_max_batched_writeback; /* Tunable */
+	u32 nr_max_batched_writeback_copied;
 
 	struct rb_root writeback_tree;
 
@@ -423,6 +418,7 @@ struct wb_device {
 
 	struct task_struct *writeback_modulator;
 	u8 writeback_threshold; /* Tunable */
+	u8 writeback_threshold_copied;
 
 	/*--------------------------------------------------------------------*/
 
@@ -432,6 +428,7 @@ struct wb_device {
 
 	struct task_struct *sb_record_updater;
 	unsigned long update_sb_record_interval; /* Tunable */
+	unsigned long update_sb_record_interval_copied;
 
 	/*--------------------------------------------------------------------*/
 
@@ -441,6 +438,7 @@ struct wb_device {
 
 	struct task_struct *data_synchronizer;
 	unsigned long sync_data_interval; /* Tunable */
+	unsigned long sync_data_interval_copied;
 
 	/*--------------------------------------------------------------------*/
 
@@ -451,6 +449,7 @@ struct wb_device {
 	struct work_struct read_cache_work;
 	struct read_cache_cells *read_cache_cells;
 	u32 read_cache_threshold; /* Tunable */
+	u32 read_cache_threshold_copied;
 
 	/*--------------------------------------------------------------------*/
 
@@ -468,6 +467,11 @@ struct wb_device {
 
 /*----------------------------------------------------------------------------*/
 
+struct write_io {
+	void *data; /* 4KB */
+	u8 data_bits;
+};
+
 void acquire_new_seg(struct wb_device *, u64 id);
 void cursor_init(struct wb_device *);
 void flush_current_buffer(struct wb_device *);
@@ -475,7 +479,7 @@ void inc_nr_dirty_caches(struct wb_device *);
 void dec_nr_dirty_caches(struct wb_device *);
 bool mark_clean_mb(struct wb_device *, struct metablock *);
 struct dirtiness read_mb_dirtiness(struct wb_device *, struct segment_header *, struct metablock *);
-void prepare_overwrite(struct wb_device *, struct segment_header *, struct metablock *old_mb, bool overwrite_fullsize);
+int prepare_overwrite(struct wb_device *, struct segment_header *, struct metablock *old_mb, struct write_io *, u8 overwrite_bits);
 
 /*----------------------------------------------------------------------------*/
 
@@ -500,46 +504,4 @@ sector_t dm_devsize(struct dm_dev *);
 
 /*----------------------------------------------------------------------------*/
 
-/*
- * Device blockup (Marking the device as dead)
- * -------------------------------------------
- *
- * I/O error on cache device blocks up the whole system.
- * After the system is blocked up, cache device is dead, all I/Os to cache
- * device are ignored as if it becomes /dev/null.
- */
-#define mark_dead(wb) set_bit(WB_DEAD, &wb->flags)
-#define is_live(wb) likely(!test_bit(WB_DEAD, &wb->flags))
-
-/*
- * This macro wraps I/Os to cache device to add context of failure.
- */
-#define maybe_IO(proc) \
-	do { \
-		r = 0; \
-		if (is_live(wb)) {\
-			r = proc; \
-		} else { \
-			r = -EIO; \
-			break; \
-		} \
-		\
-		if (r == -EIO) { \
-			mark_dead(wb); \
-			DMERR("device is marked as dead"); \
-			break; \
-		} else if (r == -ENOMEM) { \
-			DMERR("I/O failed by ENOMEM"); \
-			schedule_timeout_interruptible(msecs_to_jiffies(1000));\
-			continue; \
-		} else if (r == -EOPNOTSUPP) { \
-			break; \
-		} else if (r) { \
-			WARN_ONCE(1, "I/O failed for unknown reason err(%d)", r); \
-			break; \
-		} \
-	} while (r)
-
-/*----------------------------------------------------------------------------*/
-
 #endif
